"""
Helpers are useful classes that help you interact with the API more easily.

.. warning::
    Unless noted, helpers are not made to be manually constructed, but
    to be used through :attr:`~aiocvv.client.ClassevivaClient.me`.
"""

from .calendar.core import Calendar, Period
from .noticeboard import (  # pylint: disable=reimported
    MyNoticeboard,
    MyNoticeboard as Noticeboard,
    File,
    Attachment,
    NoticeboardItem,
)



"""
This module contains all the enums used in the aiocvv package,
which are used for type hinting and for data that is always the same.
"""

# pylint: disable=invalid-name
from enum import Enum


class Weekday(Enum):
    """
    Represents the weekdays returned by the
    Classeviva API from the calendar endpoint.

    :ivar sunday: Sunday.
    :ivar monday: Monday.
    :ivar tuesday: Tuesday.
    :ivar wednesday: Wednesday.
    :ivar thursday: Thursday.
    :ivar friday: Friday.
    :ivar saturday: Saturday.
    """

    sunday = 1
    monday = 2
    tuesday = 3
    wednesday = 4
    thursday = 5
    friday = 6
    saturday = 7


class AbsenceCode(Enum):
    """
    Represents the absence codes.

    :ivar absence: You were absent.
    :ivar delay: You entered school late.
    :ivar short_delay: You entered school a little late.
    :ivar exit: You left school early.
    """

    absence = "ABA0"
    delay = "ABR0"
    short_delay = "ABR1"
    exit = "ABU0"


class SchoolDayStatus(Enum):
    """
    Represents the school day statuses.

    :ivar school: Normal school, working day.
    :ivar no_lesson: Day without lessons.
    :ivar holiday: Holiday.
    :ivar nonworking: Non-working day.
    :ivar undefined: Anything else?
    """

    school = "SD"
    no_lesson = "ND"
    holiday = "HD"
    nonworking = "NW"
    undefined = "US"


class EventCode(Enum):
    """
    Represents the event types for agenda.

    :ivar note: An annotation for that moment in the agenda.
    :ivar homework: An assignment for that day.
    :ivar reservation: A classroom reservation.
    """

    note = "AGNT"
    homework = "AGHW"
    reservation = "AGCR"


class LessonEvent(Enum):
    """
    Represents the lesson types.

    :ivar register: Normal lesson.
    :ivar co_presense: Co-presence lesson (there is more than one teacher).
    :ivar co_presense_support: Co-presence lesson with support.
    """

    register = "LSF0"
    co_presense = "LSC0"
    co_presense_support = "LSS0"


class LessonStatus(Enum):
    """
    Represents the student's presence status for a single lesson.

    :ivar present: The student was present.
    :ivar out: The student was out of class, but present in school.
    :ivar absent: The student was absent.
    :ivar no_lesson: There was no lesson at all.
    """

    present = "HAT0"
    out = "HAT1"
    absent = "HAB0"
    no_lesson = "HNN0"


class NoteType(Enum):
    """
    Represents the note types.

    :ivar teacher: An annotation made by a teacher.
    :ivar registry: A disciplinary note.
    :ivar warning: A warning note.
    :ivar sanction: A disciplinary sanction.
    """

    teacher = "NTTE"
    registry = "NTCL"
    warning = "NTWN"
    sanction = "NTST"


class RegisterType(Enum):
    """
    Represents the register types.

    :ivar standard: Standard register.
    :ivar simplified: Simplified register.
    """

    standard = "STD"
    simplified = "SMART"


class UserType(Enum):
    """
    Represents the user types.

    :ivar student: The user is a student.
    :ivar teacher: The user is a teacher.
    :ivar parent: The user is a parent.
    """

    student = "S"
    teacher = ""  # TODO: find out what this is
    parent = "G"


class GradeCode(Enum):
    """
    Represents the different grade types.

    :ivar decimal: A simple, normal grade.
    :ivar competences: A grade for competences.
    :ivar new_competences: A grade for new competences.
    :ivar entry_test: An entry test grade.
    :ivar pcto: A grade for the PCTO.
    """

    decimal = "GRV0"
    competences = "GRV1"
    new_competences = "GRV2"
    entry_test = "GRT1"
    pcto = "GRA1"



"""
This helper contains the Calendar class, which is used to
interact with the calendar endpoint of the Classeviva API.

This class is used to get information about the user's calendar,
such as school days, absences, events, grades, notes, and more.
"""

from typing import Optional, List, AsyncIterator
from datetime import datetime, timedelta, date
from ...utils import parse_date, group_by_date, create_repr
from ...modules import StudentsModule
from ...types import Date
from ...dataclasses import (
    SchoolDay,
    AbsenceDay,
    Event,
    AgendaDay,
    Subject,
    PartialSubject,
    Lesson,
    Day,
)
from ...enums import (
    SchoolDayStatus,
    AbsenceCode,
    EventCode,
    LessonEvent,
    LessonStatus,
    NoteType,
    Weekday,
)
from .period import Period


class Calendar:
    """
    Represents the whole calendar of a student.
    """

    def __init__(
        self, module: StudentsModule, id: int
    ):  # pylint: disable=redefined-builtin
        self.module = module
        self.id = id

    @staticmethod
    def __dateify(string: str):
        return datetime.strptime(string, "%Y-%m-%d").date()

    @staticmethod
    def __timeify(string: str):
        return datetime.strptime(string, "%Y-%m-%dT%H:%M:%S%z")

    @classmethod
    def __parse_lesson(cls, data: dict) -> List[Lesson]:
        return Lesson(
            id=data["evtId"],
            date=cls.__dateify(data["evtDate"]),
            type=LessonEvent(data["evtCode"]),
            position=data["evtHPos"],
            duration=data["evtDuration"],
            class_desc=data["classDesc"],
            subject=PartialSubject(data["subjectCode"], data["subjectDesc"]),
            status=LessonStatus(data["status"]) if "status" in data else None,
        )

    @classmethod
    def __parse_event(cls, ev: dict, subjects: List[Subject]):
        return Event(
            id=ev["evtId"],
            type=EventCode(ev["evtCode"]),
            start=cls.__timeify(ev["evtDatetimeBegin"]),
            end=cls.__timeify(ev["evtDatetimeEnd"]),
            full_day=ev["isFullDay"],
            notes=ev["notes"],
            author=ev["authorName"],
            class_desc=ev["classDesc"],
            subject=(
                list(filter(lambda s: s.id == ev["subjectId"], subjects))[0]
                if ev["subjectId"]
                else None
            ),
            homework=ev["homeworkId"],
            homework_item=ev.get("homeworkItem", None),
        )

    @classmethod
    def __parse_absence(cls, evt: dict):
        return AbsenceDay(
            id=evt["evtId"],
            type=AbsenceCode(evt["evtCode"]),
            date=cls.__dateify(evt["evtDate"]),
            justified=evt["isJustified"],
            position=evt["evtHPos"],
        )

    @staticmethod
    def __parse_school_day(data: dict):
        return SchoolDay(
            date=data["dayDate"],
            weekday=data["dayOfWeek"],
            status=SchoolDayStatus(data["dayStatus"]),
        )

    async def get_school_days(
        self, begin: Optional[Date] = None, end: Optional[Date] = None
    ) -> List[SchoolDay]:
        """
        Get the school days in a range of dates.

        :param begin: The start date.
        :param end: The end date.
        :return: A list of :class:`~aiocvv.dataclasses.SchoolDay` objects.
        """
        ret = (await self.module.calendar(self.id, begin, end))["content"]
        return [self.__parse_school_day(day) for day in ret["calendar"]]

    async def get_absences(
        self, begin: Optional[Date] = None, end: Optional[Date] = None
    ) -> List[AbsenceDay]:
        """
        Get the absences in a range of dates.

        :param begin: The start date.
        :param end: The end date.
        :return: A list of :class:`~aiocvv.dataclasses.AbsenceDay` objects.
        """
        ret = await self.module.absences(self.id, begin, end)
        return [
            AbsenceDay(
                id=evt["evtId"],
                type=AbsenceCode(evt["evtCode"]),
                date=parse_date(evt["evtDate"]),
                justified=evt["isJustified"],
                position=evt["evtHPos"],
            )
            for evt in ret["content"]["events"]
        ]

    async def get_agenda(
        self,
        begin: Date,
        end: Date,
        event_code: Optional[EventCode] = None,
        *,
        separate_days: bool = True,
    ) -> AgendaDay:
        """
        Get the agenda in a range of dates.

        :param begin: The start date.
        :param end: The end date.
        :param event_code: The event code to filter by.
        :param separate_days: Whether to separate the events by day.
        :return: A list of :class:`~aiocvv.dataclasses.AgendaDay` objects.
        """
        ret = await self.module.agenda(self.id, begin, end, event_code)
        subjects = await self.module.client.me.get_subjects()

        if not separate_days:
            return [
                self.__parse_event(evt, subjects) for evt in ret["content"]["agenda"]
            ]

        days = group_by_date(ret["content"]["agenda"], self.__parse_event, subjects)

        return [AgendaDay(date, events) for date, events in days.items()]

    async def get_lessons(
        self,
        begin: Date,
        end: Optional[Date] = None,
        *,
        subject: Optional[int] = None,
    ) -> List[Lesson]:
        """
        Get the lessons in a range of dates.

        :param begin: The start date.
        :param end: The end date.
        :param subject: The subject to filter by.
        :return: A list of :class:`~aiocvv.dataclasses.Lesson` objects.
        """

        ret = await self.module.lessons(self.id, begin, end, subject=subject)
        return [self.__parse_lesson(l) for l in ret["content"]["lessons"]]

    async def get_periods(self):
        """
        Get the periods of the student's school.

        :return: A list of :class:`~aiocvv.dataclasses.Period` objects.
        """

        ret = await self.module.periods(self.id)
        return [Period(self, **period) for period in ret["content"]["periods"]]

    async def get_day(self, start: Date, end: Optional[Date] = None):
        """
        Get all of the information available for a specific day,
        merging calendar, absences, agenda, grades, notes, and
        school days all together here.

        :param start: The date of the day.
        :param end: The end date.
        :return: A list of :class:`~aiocvv.dataclasses.Day` objects.
        """

        subjects = await self.module.client.me.get_subjects()
        periods = await self.get_periods()
        schooldays = await self.module.calendar(self.id, start, end or start)
        schooldays = schooldays["content"]
        return await self.__do_get_day(subjects, periods, schooldays, start, end)

    async def __do_get_day(
        self,
        subjects,
        periods,
        schooldays,
        start: Date,
        end: Optional[Date] = None,
    ) -> List[Day]:
        me = self.module.client.me
        data = await self.module.overview(self.id, start, end)
        data = data["content"]

        lessons = group_by_date(data["lessons"], self.__parse_lesson)
        agenda = group_by_date(data["agenda"], self.__parse_event, subjects)
        events = group_by_date(data["events"], self.__parse_absence)
        grades = group_by_date(
            data["grades"], me._parse_grade, subjects, periods
        )  # pylint: disable=protected-access
        schooldays = group_by_date(schooldays["calendar"])
        notes = {}
        for tp in NoteType:
            notes.update(
                group_by_date(data["notes"][tp.value], me._parse_note, tp)
            )  # pylint: disable=protected-access

        # merge all the days together without duplicates
        days = list(
            set(
                list(lessons.keys())
                + list(agenda.keys())
                + list(events.keys())
                + list(grades.keys())
                + list(notes.keys())
                + list(schooldays.keys())
            )
        )
        ret = []
        for d in days:
            try:
                schoolday = schooldays[d][0]
            except KeyError:
                # this day shouldn't even be returned, let's ignore it
                continue

            ret.append(
                Day(
                    date=d,
                    weekday=Weekday(schoolday["dayOfWeek"]),
                    status=SchoolDayStatus(schoolday["dayStatus"]),
                    lessons=lessons.get(d, []),
                    agenda=agenda.get(d, []),
                    events=events.get(d, []),
                    grades=grades.get(d, []),
                    notes=notes.get(d, []),
                )
            )

        return ret

    @staticmethod
    def __filter_check(begin: Date, day: int):
        def chk(d: Day):
            return d.date == (begin + timedelta(days=day)).date()

        return chk

    async def __call__(self, begin: Date, end: Date) -> AsyncIterator[Day]:
        """
        Iterate over the days in a range of dates, returning
        the information available for each day, merging calendar,
        absences, agenda, grades, notes, and school days all together here.
        """
        if end < begin:
            raise ValueError("end date cannot be before begin date")

        subjects = await self.module.client.me.get_subjects()
        periods = await self.get_periods()
        schooldays = await self.module.calendar(self.id, begin, end or begin)
        schooldays = schooldays["content"]
        for day in range((end - begin).days + 1):
            yield list(
                filter(
                    self.__filter_check(begin, day),
                    await self.__do_get_day(subjects, periods, schooldays, begin, end),
                )
            )[0]



"""
This helper contains the Period class, which is used to
get information about school periods and everything that
happened during them.
"""

from ...utils import parse_date, create_repr
from ...dataclasses import Grade, Subject, Note, AbsenceDay
from ...enums import NoteType
from ...types import date
from typing import List, Optional
from typing_extensions import Self


class Period:
    """
    Represents a school period (e.g. first quarter, second quarter, etc.).
    """

    def __init__(self, calendar, **data):

        self.__calendar = calendar
        self.__data = data

        self.__code: str = data["periodCode"]
        self.__position: int = data["periodPos"]
        self.__description: str = data["periodDesc"]
        self.__final: bool = data["isFinal"]
        self.__start: date = parse_date(data["dateStart"])
        self.__end: date = parse_date(data["dateEnd"])
        self.__miur: Optional[str] = data["miurDivisionCode"]

    @property
    def code(self) -> str:
        """The code of the period."""
        return self.__code

    @property
    def position(self) -> int:
        """The period's position."""
        return self.__position

    @property
    def description(self) -> str:
        """The description of the period."""
        return self.__description

    @property
    def final(self) -> bool:
        """Whether the period is final or not."""
        return self.__final

    @property
    def start(self) -> date:
        """The date of when the period starts."""
        return self.__start

    @property
    def end(self) -> date:
        """The date of when the period ends."""
        return self.__end

    @property
    def miur_division_code(self) -> Optional[str]:
        """The division code provided by MIUR, if any."""
        return self.__miur

    def __repr__(self):
        return create_repr(
            self,
            code=self.code,
            description=self.description,
            start=self.start,
            end=self.end,
            final=self.final,
        )

    def __str__(self):
        return self.description

    def __eq__(self, other: Self):
        return self.code == other.code

    async def get_grades(self, subject: Optional[Subject] = None) -> List[Grade]:
        """
        Get the grades that were given during the period.

        :return: A list of :class:`~aiocvv.dataclasses.Grade` objects.
        """
        resp = list(
            filter(
                lambda g: g["periodPos"] == self.position,
                (
                    await self.__calendar.module.grades(
                        self.__calendar.id, subject.id if subject else None
                    )
                )["content"]["grades"],
            )
        )

        me = self.__calendar.module.client.me
        subjects = await me.get_subjects()
        return [
            me._parse_grade(g, subjects, [self])  # pylint: disable=protected-access
            for g in resp
        ]

    async def get_notes(self, type: Optional[NoteType] = None) -> List[Note]:
        """
        Get the notes assigned during the period.

        :return: A list of :class:`~aiocvv.dataclasses.Note` objects.
        """
        ret: List[Note] = await self.__calendar.module.client.me.get_notes(type)
        return list(filter(lambda n: self.start <= n.date <= self.end, ret))

    async def get_absences(self) -> List[AbsenceDay]:
        """
        Get the days the student has been absent during this period.

        :return: A list of :class:`~aiocvv.dataclasses.AbsenceDay` objects.
        """
        return await self.__calendar.get_absences(self.start, self.end)



"""
Useful functions used inside the library.
"""

from datetime import datetime, date, timedelta
from typing import Union, Type, Callable, Optional
from .errors import ClassevivaError
from .types import AnyCVVError


def create_repr(self, **kwargs):
    """
    Create a __repr__ string for a class.
    """
    params = []
    for k, v in kwargs.items():
        if v is not None:
            if isinstance(v, list):
                v = len(v)

            params.append(f"{k}={v!r}")

    return f"<{type(self).__name__} {' '.join(params)}>"


def convert_date(date_: Union[datetime, date], today: bool = False) -> str:
    """
    Convert a date to a string.
    """
    date_ = getattr(date_, "date", lambda: date_)()
    if today and date_ in [date.today(), date.today() - timedelta(days=1)]:
        return "today" if date_ == date.today() else "yesterday"

    return date_.strftime("%Y%m%d")


def __recurse_subclasses(cls: Type):
    for sub in cls.__subclasses__():
        yield sub
        yield from __recurse_subclasses(sub)


def find_exc(
    response: dict, base: Type[ClassevivaError] = ClassevivaError
) -> AnyCVVError:
    """
    Find the correct exception to raise based
    on the response from the Classeviva API.
    """
    content = response["content"]
    tp = content["error"].split("/")[1]
    status = response["status"]
    if not issubclass(base, ClassevivaError):
        raise ValueError("base must derive from ClassevivaError")

    if tp == "authentication failed":
        sc = content["info"]
    elif status < 200 or status >= 300:
        sc = response["status_reason"].replace(" ", "")

    exc = base(response)
    for sub in __recurse_subclasses(base):
        if sub.__name__ == sc:
            exc = sub(response)
            break

    return exc


def capitalize_name(string: str):
    """
    Capitalizes a name.
    """
    return " ".join(word.capitalize() for word in string.split())


def parse_date(string: str):
    """
    Converts a date string in the YYYY-mm-dd format to a date object.
    """
    return datetime.strptime(string, "%Y-%m-%d").date()


def parse_time(string: str):
    """
    Converts a time string in the YYYY-mm-ddTHH:MM:SS+HH:MM format to a datetime object.
    """
    return datetime.strptime(string, "%Y-%m-%dT%H:%M:%S%z")


def group_by_date(
    data: list, parser: Optional[Callable] = None, *args, **kwargs
):  # pylint: disable=keyword-arg-before-vararg
    """
    Groups a list of events by date.
    """
    ret = {}
    for dt in data:
        date_ = (
            parse_time(dt["evtDatetimeBegin"]).date()
            if "evtDatetimeBegin" in dt
            else parse_date(dt["evtDate" if "evtDate" in dt else "dayDate"])
        )

        if date_ not in ret:
            ret[date_] = []

        ret[date_].append(parser(dt, *args, **kwargs) if parser else dt)

    return ret



"""An asynchronous API wrapper for Classeviva."""

from . import client
from . import errors
from . import utils
from .errors import *
from .enums import *
from .dataclasses import *
from .me import Me
from .client import ClassevivaClient
from .client import ClassevivaClient as Client  # pylint: disable=reimported

__version__ = "0.1.1"
__author__ = "Vinche.zsh"
__email__ = "vinchethescript@gmail.com"
__license__ = "GPL-3.0"
__description__ = "An API wrapper for Classeviva written in Python using asyncio."
__url__ = "https://github.com/Vinchethescript/aiocvv"



"""
This helper contains the classes that represent the noticeboard and its items.
The noticeboard can be used both by students and teachers.
"""

# pylint: disable=arguments-differ

from typing import Optional, Any, IO, List, Union
from io import BytesIO
from ..modules.core import BaseModule, Noticeboard
from ..types import Response
from ..errors import ClassevivaError


class File:
    """
    Represents a file to upload to the noticeboard when joining an item.

    This class is used as an argument for :meth:`~aiocvv.helpers.noticeboard.NoticeboardItem.join`.
    """

    def __init__(self, data: IO[Any], filename: str):
        data.seek(0)
        self.__data = data
        self.filename = filename

    @property
    def data(self) -> IO[Any]:
        """The file's data."""
        return self.__data

    def __str__(self):
        return self.filename

    def __repr__(self):
        return f"<File name={self.filename!r}>"


class PartialAttachment:
    """
    Represents an attachment of an uploaded noticeboard item.

    .. note::
        This comes from a *partial* item, so the attachment is not downloadable.
        To download the attachment, you must :func:`~aiocvv.helpers.noticeboard.PartialNoticeboardItem.read` the item first.
    """

    def __init__(self, data: dict):
        # pylint: disable=super-init-not-called
        self.filename = data["fileName"]
        self.num = data["attachNum"]


class Attachment(File, PartialAttachment):
    """
    Represents an attachment of an uploaded noticeboard item.
    """

    def __init__(self, item: "NoticeboardItem", data: dict):
        PartialAttachment.__init__(self, data)
        self.__item = item

    async def download(self):
        """Download the attachment."""
        resp = await self.__item.noticeboard.noticeboard.get_attachment(
            self.__item.noticeboard.id, self.__item.code, self.__item.id, self.num
        )
        data = BytesIO(resp["content"])
        super().__init__(data, self.filename)
        return self


class PartialNoticeboardItem:
    """
    Represents a item in the noticeboard *partially*.

    .. note::
        This class is *partial*, which means it must be :func:`~aiocvv.helpers.noticeboard.PartialNoticeboardItem.read` to get its content.
        This is because reading the item would change its read status.
    """

    def __init__(self, nb: "MyNoticeboard", payload: dict):
        self.__payload = payload
        self.__board = nb
        self.__attachments = []

    @property
    def noticeboard(self):
        """The noticeboard this item belongs to."""
        return self.__board

    @property
    def id(self) -> int:
        """The item's publication ID."""
        return self.__payload["pubId"]

    @property
    def is_read(self):
        """Whether the item has been read."""
        return self.__payload["readStatus"]

    @property
    def code(self) -> str:
        """The item's event code."""
        return self.__payload["evtCode"]

    @property
    def title(self) -> str:
        """The item's title."""
        return self.__payload["cntTitle"]

    @property
    def category(self) -> str:
        """The item's category."""
        return self.__payload["cntCategory"]

    @property
    def has_changed(self) -> bool:
        """Whether the item has changed."""
        return self.__payload["cntHasChanged"]

    @property
    def attachments(self):
        """The item's attachments."""
        if not self.__attachments:
            for attach in self.__payload["attachments"]:
                if isinstance(self, NoticeboardItem):
                    self.__attachments.append(Attachment(self, attach))
                else:
                    self.__attachments.append(PartialAttachment(attach))

            self.__attachments = list(sorted(self.__attachments, key=lambda x: x.num))

        return self.__attachments

    async def read(self) -> Response:
        """
        Read the item.

        :return: The noticeboard full item.
        """
        return await self.__board.read(self.code, self.id)


class NoticeboardItem(PartialNoticeboardItem):
    """
    Represents a item in the noticeboard. See also :class:`~aiocvv.helpers.noticeboard.PartialNoticeboardItem`.
    """

    def __init__(self, nb: "MyNoticeboard", payload: dict, content: str):
        super().__init__(nb, payload)
        self.__content = content

    @property
    def content(self):
        """The item's content."""
        return self.__content

    async def read(self):
        """
        This function doesn't do anything, but exists for compatibility with partial items and to avoid useless requests.

        See also :func:`~aiocvv.helpers.noticeboard.PartialNoticeboardItem.read`.
        :return: This item.
        """
        return self

    async def join(
        self,
        text: Optional[str] = None,
        sign: Optional[bool] = None,
        file: Optional[File] = None,
        *,
        raise_exc: bool = True,
    ) -> bool:
        """
        Join this item.

        :param text: The text to send.
        :param sign: Whether to sign the text.
        :param file: The file to upload.
        :param raise_exc: Whether to raise an exception if an error occurs.

        :return: Whether the operation was successful.
        """
        try:
            await self.__board.noticeboard.join(
                self.__payload["evtCode"],
                self.__payload["pubId"],
                text=text,
                filename=file.filename if file else None,
                file=file.data if file else None,
                sign=sign,
                attrs=False,
                include_attachment=False,
                reply_info=False,
            )
        except ClassevivaError:
            if raise_exc:
                raise

            return False

        return True


AnyNoticeboardItem = Union[PartialNoticeboardItem, NoticeboardItem]


class MyNoticeboard:
    """
    Represents the noticeboard of a user.
    """

    def __init__(
        self, noticeboard: Noticeboard, id: int
    ):  # pylint: disable=redefined-builtin

        self.noticeboard = noticeboard
        self.id = id
        self.__read = self.noticeboard.read

    async def all(self) -> List[AnyNoticeboardItem]:
        """Get all the items in the noticeboard."""
        ret = []
        async for item in self:
            ret.append(item)
        return ret

    async def __get(
        self, code: str, id: int  # pylint: disable=redefined-builtin
    ) -> Response:
        items = await self.noticeboard.all(self.id)
        for item in items["content"]["items"]:
            if id == item["pubId"] and code == item["evtCode"]:
                return item

    async def get(
        self, code: str, id: int  # pylint: disable=redefined-builtin
    ) -> AnyNoticeboardItem:
        """
        Get a noticeboard item.

        .. note::
            The returned item may be *partial*, meaning that it doesn't contain some data.
            This is because you have to read the item to get the rest of the data, and that would change the read status of the item.

        :param code: The event code of the item.
        :param id: The publication ID of the item.

        :return: The partial noticeboard item.
        """

        payload = await self.__get(code, id)
        if payload["readStatus"]:
            return await self.read(code, id)

        return PartialNoticeboardItem(self, payload)

    async def read(self, event_code: str, publication_id: int) -> NoticeboardItem:
        """
        Read a noticeboard item.

        .. note::
            This will automatically mark the item as read from the Classeviva backend.

        :param event_code: The event code of the item.
        :param publication_id: The publication ID of the item.

        :return: The noticeboard full item.
        """
        data = await self.__read(self.id, event_code, publication_id)
        payload = await self.__get(event_code, publication_id)

        return NoticeboardItem(self, payload, data["content"]["item"]["text"])

    async def __aiter__(self):
        data = await self.noticeboard.all(self.id)

        for item in data["content"]["items"]:
            # return the full item if it's already been read
            if item["readStatus"]:
                yield NoticeboardItem(
                    self,
                    item,
                    (await self.__read(self.id, item["evtCode"], item["pubId"]))[
                        "content"
                    ]["item"]["text"],
                )
            else:
                yield PartialNoticeboardItem(
                    self,
                    item,
                )



"""The module where the ClassevivaClient class is located."""

import os
import asyncio
import json
from datetime import datetime
from types import SimpleNamespace
from typing import Optional, Mapping, Any, Iterable, Union, Tuple
from urllib.parse import urljoin, urlsplit, urlparse

import aiohttp
from diskcache import Cache
from appdirs import user_cache_dir
from typing_extensions import Self
from aiohttp.client import (
    Fingerprint,
    ClientTimeout,
    SSLContext,  # to avoid checking if `ssl` module exists
)
from aiohttp.helpers import sentinel
from aiohttp.typedefs import StrOrURL, LooseCookies, LooseHeaders
from .core import CLIENT_USER_AGENT, CLIENT_DEV_APIKEY, CLIENT_CONTENT_TP
from .modules import (
    TeachersModule,
    StudentsModule,
    ParentsModule,
)
from . import me
from .errors import AuthenticationError
from .me import UserType, Teacher, Student, Parent
from .types import Response
from .utils import find_exc
from ._auth import AuthenticationModule

_json = json
LoginMethods = Union[Tuple[str, str], Tuple[str, str, str]]


class ClassevivaClient:
    """
    The client class for Classeviva.

    This class provides an interface to interact with the Classeviva REST APIs,
    where all the requests from this module are made.
    You can get any information from Classeviva by either using the :attr:`me` attribute
    or by making manual requests using the appropriate module or the :meth:`request` method-

    The modules are used to make manual HTTP requests to the Classeviva APIs.
    This is useful when:

    * You use the :attr:`me` attribute to automatically get the information you need, which chooses the correct module to use;
    * You want to make a request to an endpoint that is not (yet) or partially implemented in this wrapper;
    * You want to have more control over the request or response.
    * You don't want to manually write every URL, like you would do with the :meth:`request` method.

    .. note::
        Of course, if you decide to use modules,
        you will have to parse the response yourself,
        as they only return the raw response.


    For more information about the endpoints,
    refer to the Classeviva REST API documentation:

    * `HTML Documentation <https://web.spaggiari.eu/rest/v1/docs/html>`_
    * `Plaintext Documentation <https://web.spaggiari.eu/rest/v1/docs/plaintext>`_


    :param username: The username for authentication.
    :param password: The password for authentication.
    :param identity: Optional. The identity for authentication.
    :param loop: Optional. The event loop to use.
                 If not provided, the default event loop will be used.
    :param base_url: Optional. The base URL for the Classeviva REST APIs.
                        Default is `https://web.spaggiari.eu/rest/v1/`.
    :param strict_caching: Optional. Whether to strictly use caching.
                           Setting this to True might introduce some delays in
                           updates, but will reduce the number of requests made
                           and will make the client faster.

    :type username: str
    :type password: str
    :type identity: str
    :type loop: asyncio.AbstractEventLoop
    :type base_url: str
    """

    def __init__(
        self,
        username: str,
        password: str,
        identity: Optional[str] = None,
        *,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        base_url: str = "https://web.spaggiari.eu/rest/v1/",
        strict_caching: bool = True,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.__username = username
        self.__password = password
        self.__identity = identity
        self._cache_path = os.path.join(user_cache_dir(), "aiocvv")
        self.__auth = AuthenticationModule(
            self
        )  # NOTE: keeping this private for obvious reasons
        self.__teachers = None
        self.__students = None
        self.__parents = None
        self.__me = None
        self._base_url = base_url
        self.__parsed_base = urlparse(base_url)
        self.strict_caching = strict_caching

    @property
    def base_url(self) -> str:
        """
        The base URL for the Classeviva REST APIs.

        You can change this if:
        * You're trying to fetch data from a past school year;
        * You're testing against a local or private instance of Classeviva.

        .. note::
            If you're changing this to get data from a past school year, the
            base URL will be something like `https://webYY.spaggiari.eu/rest/v1/`,
            where `YY` is the last two digits of the year.
            For example, in August 2025, the previous school year was "2024-2025",
            so the base URL would be `https://web24.spaggiari.eu/rest/v1/`.

        :return: The base URL.
        """
        return self._base_url

    @base_url.setter
    def base_url(self, value: str):
        self._base_url = value
        self.__parsed_base = urlparse(value)

    #          Modules          #
    # Using properties here so only the needed
    # modules will be initialized when needed.

    @property
    def teachers(self) -> TeachersModule:
        """
        Get the teachers module for manually making
        requests to the teachers' endpoints.

        .. warning::
            This whole module has not been implemented yet, because it's
            not possible to test it without having a teacher account.
            Therefore, this is just a placeholder for future updates, which
            means there are no endpoints in here. If you are a teacher and
            you want to contribute to this project, feel free to open a pull request.
            Otherwise, you'll have to manually make requests using the :meth:`request` method.

        :return: The (empty) module instance.
        """
        if self.__teachers is None:
            self.__teachers = TeachersModule(self)

        return self.__teachers

    @property
    def students(self) -> StudentsModule:
        """
        Get the students module for manually making
        requests to the students' endpoints.

        :return: The module instance.
        """
        if self.__students is None:
            self.__students = StudentsModule(self)

        return self.__students

    @property
    def parents(self) -> ParentsModule:
        """
        Get the parents module for manually making
        requests to the parents' endpoints.

        :return: The module instance.
        """
        if self.__parents is None:
            self.__parents = ParentsModule(self)

        return self.__parents

    @property
    def me(self) -> Optional[Union[Student, Parent, Teacher]]:
        """
        Get the current user.

        .. note::
            This will be None until :meth:`login` is called or the class is awaited.

        :return: The current user instance.
        """
        return self.__me

    async def request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Any = None,
        json: Optional[dict] = None,  # pylint: disable=redefined-outer-name
        cookies: Optional[LooseCookies] = None,
        headers: Optional[LooseHeaders] = None,
        skip_auto_headers: Optional[Iterable[str]] = None,
        compress: Optional[str] = None,
        chunked: Optional[bool] = None,
        raise_for_status: bool = True,
        read_until_eof: bool = True,
        proxy: Optional[StrOrURL] = None,
        timeout: Union[ClientTimeout, object] = sentinel,
        verify_ssl: Optional[bool] = None,
        fingerprint: Optional[bytes] = None,
        ssl_context: Optional[SSLContext] = None,
        ssl: Optional[Union[SSLContext, bool, Fingerprint]] = None,
        proxy_headers: Optional[LooseHeaders] = None,
        trace_request_ctx: Optional[SimpleNamespace] = None,
        read_bufsize: Optional[int] = None,
    ) -> Response:
        """
        Make a raw HTTP request to the Classeviva REST APIs using aiohttp.

        For information about the valid endpoints,
        refer to the Classeviva REST API documentation:

        * `HTML Documentation <https://web.spaggiari.eu/rest/v1/docs/html>`_
        * `Plaintext Documentation <https://web.spaggiari.eu/rest/v1/docs/plaintext>`_

        :param method: The HTTP method to use.
        :param endpoint: The path for the request.
                         Must be a relative URL since the base URL is `https://web.spaggiari.eu/rest/v1/`.
        :param params: Optional. The query parameters for the request.
        :param data: Optional. The request body data.
        :param json: Optional. The request body JSON data.
        :param cookies: Optional. The cookies to include in the request.
        :param headers: Optional. The headers to include in the request.
        :param skip_auto_headers: Optional. The headers to skip from automatic inclusion.
        :param compress: Optional. The compression method to use.
        :param chunked: Optional. Whether to use chunked transfer encoding.
        :param raise_for_status: Optional. Whether to raise an exception for non-successful responses.
        :param read_until_eof: Optional. Whether to read the response until EOF.
        :param proxy: Optional. The proxy URL or path.
        :param timeout: Optional. The request timeout.
        :param verify_ssl: Optional. Whether to verify SSL certificates.
        :param fingerprint: Optional. The SSL fingerprint.
        :param ssl_context: Optional. The SSL context.
        :param ssl: Optional. The SSL configuration.
        :param proxy_headers: Optional. The headers to include in the proxy request.
        :param trace_request_ctx: Optional. The request context for tracing.
        :param read_bufsize: Optional. The read buffer size.

        :type method: str
        :type endpoint: str
        :type params: Optional[Mapping[str, str]]
        :type data: Any
        :type json: dict
        :type cookies: Optional[LooseCookies]
        :type headers: Optional[LooseHeaders]
        :type skip_auto_headers: Optional[Iterable[str]]
        :type compress: Optional[str]
        :type chunked: Optional[bool]
        :type raise_for_status: bool
        :type read_until_eof: bool
        :type proxy: Optional[StrOrURL]
        :type timeout: ClientTimeout
        :type verify_ssl: Optional[bool]
        :type fingerprint: Optional[bytes]
        :type ssl_context: Optional[SSLContext]
        :type ssl: Optional[Union[SSLContext, bool, Fingerprint]]
        :type proxy_headers: Optional[LooseHeaders]
        :type trace_request_ctx: Optional[SimpleNamespace]
        :type read_bufsize: Optional[int]

        :return: The HTTP response dictionary.
        """
        if urlsplit(endpoint).scheme:
            raise ValueError(
                f"Invalid URL given: The URL provided is not for {self.base_url}."
            )

        if not endpoint.startswith(self.base_url):
            endpoint = urljoin(self.base_url, endpoint.lstrip("/"))

        login = await self.__auth.login(
            self.__username, self.__password, self.__identity
        )
        token = login["token"]

        parsed_url = urlparse(endpoint)
        cache_ = await self.loop.run_in_executor(None, Cache, self._cache_path)
        if self.base_url not in cache_:
            cache_[self.base_url] = {}

        cache = cache_[self.base_url]

        if "requests" not in cache:
            cache["requests"] = {}

        reqs_cache = cache["requests"]
        part = parsed_url.path[len(self.__parsed_base.path) :]

        try:
            _headers = {
                "User-Agent": CLIENT_USER_AGENT,
                "Z-Dev-Apikey": CLIENT_DEV_APIKEY,
                "Content-Type": CLIENT_CONTENT_TP,
                "Z-Auth-Token": token,
            }

            if headers:
                headers.update(_headers)
            else:
                headers = _headers

            if part in reqs_cache:
                headers["If-None-Match"] = reqs_cache[part]["etag"]
                old_req_headers = reqs_cache[part]["headers"]
                headers_lower = {k.lower(): v for k, v in old_req_headers.items()}
                if "z-cache-control" in headers_lower:
                    for val in headers_lower["z-cache-control"].split(","):
                        k, v = val.strip().split("=")
                        if k.strip() == "max-age":
                            expires_at = datetime.fromtimestamp(
                                reqs_cache[part]["created_at"] + int(v.strip(" ;"))
                            )
                            if expires_at > datetime.now():
                                resp = reqs_cache[part]
                                if raise_for_status and (
                                    resp["status"] < 200 or resp["status"] >= 300
                                ):
                                    raise find_exc(resp)

                                return resp

            async with aiohttp.ClientSession() as session:
                async with session.request(
                    method,
                    endpoint,
                    params=params,
                    data=data,
                    json=json,
                    cookies=cookies,
                    headers=headers,
                    skip_auto_headers=skip_auto_headers,
                    compress=compress,
                    chunked=chunked,
                    raise_for_status=False,
                    read_until_eof=read_until_eof,
                    proxy=proxy,
                    timeout=timeout,
                    verify_ssl=verify_ssl,
                    fingerprint=fingerprint,
                    ssl_context=ssl_context,
                    ssl=ssl,
                    proxy_headers=proxy_headers,
                    trace_request_ctx=trace_request_ctx,
                    read_bufsize=read_bufsize,
                ) as resp:
                    if resp.status == 304:
                        if self.strict_caching:
                            # keep this cached for longer until it expires again
                            reqs_cache[part]["created_at"] = datetime.now().timestamp()

                        return reqs_cache[part]

                    read_data = await resp.content.read()
                    content = read_data

                    try:
                        content = _json.loads(content)
                    except (_json.JSONDecodeError, UnicodeDecodeError) as e:
                        # this is not JSON
                        if not isinstance(e, UnicodeDecodeError):
                            content = read_data.decode()

                    etag = resp.headers.get("ETag")
                    ret = {
                        "created_at": datetime.now().timestamp(),
                        "content": content,
                        "headers": dict(resp.headers),
                        "status": resp.status,
                        "status_reason": resp.reason,
                    }
                    if etag:
                        ret["etag"] = etag
                        reqs_cache[part] = ret

                    if raise_for_status and (resp.status < 200 or resp.status >= 300):
                        raise find_exc(ret)

                    return ret
        finally:
            cache_[self.base_url] = cache
            await self.loop.run_in_executor(None, cache_.close)

    async def login(self, raise_exceptions: bool = True):
        """
        Log in to Classeviva using the passed credentials.

        :param raise_exceptions: Optional. Whether to raise exceptions for authentication errors.
        :type raise_exceptions: bool

        :return: True if login is successful, False otherwise.
        :rtype: bool
        """
        try:
            data = await self.__auth.login(
                self.__username, self.__password, self.__identity
            )
        except AuthenticationError:
            if raise_exceptions:
                raise

            return False

        status = await self.__auth.status(data["token"])
        utype = UserType(status["ident"][:1])

        # this is because parents doesn't have card, and
        # parents can request all of the students' endpoints
        carder = (
            self.students
            if utype == UserType.parent
            else getattr(self, utype.name + "s")
        )

        card = await carder.request(
            "GET", f'{"".join(filter(str.isdigit, status["ident"]))}/card'
        )
        cls = getattr(me, utype.name.capitalize())
        self.__me = cls(self, **card["content"]["card"])
        return True

    async def __await_login(self) -> Self:
        await self.login()
        return self

    def __await__(self):
        return self.__await_login().__await__()



"""
This module contains the base classes for the students, parents and teachers modules.
"""

import os
from abc import ABC
from types import SimpleNamespace
from typing import Optional, Mapping, Any, Iterable, Union, IO, TYPE_CHECKING
from io import BytesIO, StringIO
from base64 import b64encode
from urllib.parse import urljoin

from diskcache import Cache
from aiohttp.client import (
    Fingerprint,
    ClientTimeout,
    SSLContext,
)
from aiohttp.helpers import sentinel
from aiohttp.typedefs import StrOrURL, LooseCookies, LooseHeaders
from ..types import Response


class Noticeboard:
    """
    Represents the noticeboard functionality in the ClasseViva API.
    """

    def __init__(self, module: "Module"):
        self.module = module

    @staticmethod
    def __get_flags(*args: bool):
        return "".join(str(int(arg)) for arg in args)

    async def all(self, id: int) -> Response:  # pylint: disable=redefined-builtin
        """
        Get all the noticeboard items.

        :param id: The ID of the student/teacher.

        :return: The response from the API.
        """

        return await self.module.request("GET", f"/{id}/noticeboard")

    async def read(
        self,
        id: int,  # pylint: disable=redefined-builtin
        event_code: int,
        publication_id: int,
        *,
        attrs: bool = True,
        include_attachment: bool = False,
        reply_info: bool = False,
        multi: bool = False,
    ) -> Response:
        """
        Read a noticeboard item.

        .. note::
            This will automatically mark the item as read from the Classeviva backend.

        :param id: The ID of the student/teacher.
        :param event_code: The code of the notice.
        :param publication_id: The ID itself of the notice.
        :param attrs: Optional. Whether to include attributes.
        :param include_attachment: Optional. Whether to include attachments.
        :param reply_info: Optional. Whether to include reply information.
        :param multi: Optional. Whether to get all of the attachments of the
                      item. This is useful when ``include_attachment`` is True.

        :return: The response from the API.
        """

        flags = self.__get_flags(attrs, include_attachment, reply_info)
        multi = "multi" if multi else ""
        return await self.module.request(
            "POST",
            f"/{id}/noticeboard/read{multi}/{event_code}/{publication_id}/{flags}",
            json={"join": False},
        )

    async def __pre_join(
        self,
        *,
        text: Optional[str] = None,
        filename: Optional[str] = None,
        file: Optional[IO[Any]] = None,
        sign: Optional[bool] = None,
        attrs: bool = True,
        include_attachment: bool = False,
        reply_info: bool = False,
    ):
        flags = self.__get_flags(attrs, include_attachment, reply_info)
        payload = {"join": True}
        if text:
            payload["text"] = text

        if file:
            if isinstance(file, (BytesIO, StringIO)):
                cont = file.getvalue()
            else:
                cont = file.read()

            if not filename and not file.name:
                raise ValueError("a file name must be specified")

            payload["file"] = await self.module.client.loop.run_in_executor(
                None, lambda: b64encode(cont).decode("utf-8")
            )
            payload["filename"] = os.path.basename(
                getattr(file, "name", None) or filename
            )

        if sign:
            payload["sign"] = sign

        return flags, payload

    async def join(
        self,
        id: int,  # pylint: disable=redefined-builtin
        event_code: int,
        publication_id: int,
        *,
        text: Optional[str] = None,
        file: Optional[IO[Any]] = None,
        filename: Optional[str] = None,
        sign: Optional[bool] = None,
        attrs: bool = True,
        include_attachment: bool = False,
        reply_info: bool = False,
        multi: bool = False,
    ) -> Response:
        """
        Join a noticeboard item.

        .. note::
            This will automatically mark the item as read from the Classeviva backend.

        :param id: The ID of the student/teacher.
        :param event_code: The code of the notice.
        :param publication_id: The ID itself of the notice.
        :param text: Optional. The text to join with.
        :param file: Optional. The file object to join with.
        :param filename: Optional. The name of the file to join with.
        :param sign: Optional. Whether to sign the join.
        :param attrs: Optional. Whether to include attributes.
        :param include_attachment: Optional. Whether to include attachments.
        :param reply_info: Optional. Whether to include reply information.
        :param multi: Optional. Whether to get all of the attachments of the
                      item. This is useful when ``include_attachment`` is True.


        :return: The response from the API.
        """

        flags, payload = await self.__pre_join(
            text=text,
            filename=filename,
            file=file,
            sign=sign,
            attrs=attrs,
            include_attachment=include_attachment,
            reply_info=reply_info,
        )
        multi = "multi" if multi else ""
        return await self.module.request(
            "POST",
            f"/{id}/noticeboard/read{multi}/{event_code}/{publication_id}/{flags}",
            json=payload,
        )

    async def get_attachment(
        self,
        id: int,  # pylint: disable=redefined-builtin
        event_code: int,
        publication_id: int,
        attach_num: int = 1,
    ) -> Response:
        """
        Get an attachment from a noticeboard item.

        :param id: The ID of the student/teacher.
        :param event_code: The code of the notice.
        :param publication_id: The ID itself of the notice.
        :param attach_num: Optional. The attachment number.

        :return: The response from the API.
        """

        return await self.module.request(
            "GET",
            f"/{id}/noticeboard/attach/{event_code}/{publication_id}/{attach_num}",
        )


class Module(ABC):
    """
    This is the base module for the students, parents and teachers modules.
    It implements the basic methods that are shared between the three modules.
    """

    endpoint = None

    def __init__(self, client):
        if not self.endpoint:
            raise ValueError("An endpoint must be added.")

        self.client = client

        if TYPE_CHECKING:
            from ..client import (
                ClassevivaClient,
            )  # pylint: disable=import-outside-toplevel

            self.client: ClassevivaClient = client

    def get_cache(self) -> Cache:
        return Cache(self.client._cache_path)  # pylint: disable=protected-access

    async def request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Any = None,
        json: Any = None,
        cookies: Optional[LooseCookies] = None,
        headers: Optional[LooseHeaders] = None,
        skip_auto_headers: Optional[Iterable[str]] = None,
        compress: Optional[str] = None,
        chunked: Optional[bool] = None,
        raise_for_status: bool = True,
        read_until_eof: bool = True,
        proxy: Optional[StrOrURL] = None,
        timeout: Union[ClientTimeout, object] = sentinel,
        verify_ssl: Optional[bool] = None,
        fingerprint: Optional[bytes] = None,
        ssl_context: Optional[SSLContext] = None,
        ssl: Optional[Union[SSLContext, bool, Fingerprint]] = None,
        proxy_headers: Optional[LooseHeaders] = None,
        trace_request_ctx: Optional[SimpleNamespace] = None,
        read_bufsize: Optional[int] = None,
    ) -> Response:
        """
        Make a raw HTTP request to the Classeviva REST APIs using aiohttp.

        This function calls :meth:`ClassevivaClient.request` to make the
        request itself, the only difference is that the request will be
        forced to be inside the module.

        .. note::
            For example, if you want to make a request to
            the ``/students/<sid>/homeworks`` endpoint from the students module,
            the `endpoint` parameter should be ``/<sid>/homeworks``, otherwise
            the request will fail as it'll make a request to
            ``/students/students/<sid>/homeworks``, which is not valid.

        For information about the valid endpoints,
        refer to the Classeviva REST API documentation:

        * `HTML Documentation <https://web.spaggiari.eu/rest/v1/docs/html>`_
        * `Plaintext Documentation <https://web.spaggiari.eu/rest/v1/docs/plaintext>`_

        :param method: The HTTP method to use.
        :param endpoint: The path for the request. Must be a relative URL
                         since the base URL is `https://web.spaggiari.eu/rest/v1/<module>/`.
        :param params: Optional. The query parameters for the request.
        :param data: Optional. The request body data.
        :param json: Optional. The request body JSON data.
        :param cookies: Optional. The cookies to include in the request.
        :param headers: Optional. The headers to include in the request.
        :param skip_auto_headers: Optional. The headers to skip from automatic inclusion.
        :param compress: Optional. The compression method to use.
        :param chunked: Optional. Whether to use chunked transfer encoding.
        :param raise_for_status: Optional. Whether to raise an exception
                                 for non-successful responses.
        :param read_until_eof: Optional. Whether to read the response until EOF.
        :param proxy: Optional. The proxy URL or path.
        :param timeout: Optional. The request timeout.
        :param verify_ssl: Optional. Whether to verify SSL certificates.
        :param fingerprint: Optional. The SSL fingerprint.
        :param ssl_context: Optional. The SSL context.
        :param ssl: Optional. The SSL configuration.
        :param proxy_headers: Optional. The headers to include in the proxy request.
        :param trace_request_ctx: Optional. The request context for tracing.
        :param read_bufsize: Optional. The read buffer size.

        :type method: str
        :type endpoint: str
        :type params: Optional[Mapping[str, str]]
        :type data: Any
        :type json: dict
        :type cookies: Optional[LooseCookies]
        :type headers: Optional[LooseHeaders]
        :type skip_auto_headers: Optional[Iterable[str]]
        :type compress: Optional[str]
        :type chunked: Optional[bool]
        :type raise_for_status: bool
        :type read_until_eof: bool
        :type proxy: Optional[StrOrURL]
        :type timeout: ClientTimeout
        :type verify_ssl: Optional[bool]
        :type fingerprint: Optional[bytes]
        :type ssl_context: Optional[SSLContext]
        :type ssl: Optional[Union[SSLContext, bool, Fingerprint]]
        :type proxy_headers: Optional[LooseHeaders]
        :type trace_request_ctx: Optional[SimpleNamespace]
        :type read_bufsize: Optional[int]

        :return: The HTTP response dictionary.
        :rtype: dict
        """
        return await self.client.request(
            method,
            urljoin(
                self.endpoint.strip("/") + "/",
                endpoint.lstrip("/"),
            ),
            params=params,
            data=data,
            json=json,
            cookies=cookies,
            headers=headers,
            skip_auto_headers=skip_auto_headers,
            compress=compress,
            chunked=chunked,
            raise_for_status=raise_for_status,
            read_until_eof=read_until_eof,
            proxy=proxy,
            timeout=timeout,
            verify_ssl=verify_ssl,
            fingerprint=fingerprint,
            ssl_context=ssl_context,
            ssl=ssl,
            proxy_headers=proxy_headers,
            trace_request_ctx=trace_request_ctx,
            read_bufsize=read_bufsize,
        )


class BaseModule(Module, ABC):
    """
    Base class for student and peacher modules.
    It's not the base class for the parent module because
    it has totally different endpoints, which is why
    parents can access all of the student's endpoints.
    """

    async def get_card(self, id: int) -> Response:  # pylint: disable=redefined-builtin
        """
        Get the card of the user.
        """
        return await self.request("GET", f"/{id}/card")

    @property
    def noticeboard(self) -> Noticeboard:
        """
        Get the noticeboard-related endpoints, which have
        been separated to avoid making too many functions
        in one class for the same endpoints.
        """
        return Noticeboard(self)



"""
This is where the dataclasses used to represent the Classeviva API responses are defined.

They are used to represent the data returned by the API in a more structured and easy-to-use way.

.. warning:: 
    These should **NOT** be manually constructed, as they are returned
    by the methods from :attr:`~aiocvv.client.ClassevivaClient.me`.

    In this case, parameters should be considered as attributes, and not as parameters.
    Dataclasses are frozen, which means that they're immutable and cannot be changed after creation.
"""

from dataclasses import dataclass
from typing import Optional, List
from datetime import date, datetime
from .enums import (
    AbsenceCode,
    Weekday,
    SchoolDayStatus,
    EventCode,
    LessonEvent,
    LessonStatus,
    NoteType,
    GradeCode,
)
from .utils import create_repr


@dataclass(frozen=True)
class AbsenceDay:
    """
    Represents an absence day.
    This could be either a full-day absence, a (small) delay or an early exit.

    :param id: The ID of the absence.
    :param type: The type of absence.
    :param date: The day of the absence.
    :param justified: Whether the absence is justified.
    :param position: The position of the delay or exit.
    """

    id: int
    type: AbsenceCode
    date: date
    justified: bool
    position: Optional[int]  # only if type is delay or exit

    def __int__(self):
        return self.id

    def __repr__(self) -> str:
        return create_repr(
            self, id=self.id, date=self.date, type=self.type, justified=self.justified
        )


@dataclass(frozen=True)
class SchoolDay:
    """
    Represents a school day.

    :param date: The date.
    :param weekday: The day weekday.
    :param status: The kind of day (normal, non-working, etc.).
    """

    date: date
    weekday: Weekday
    status: SchoolDayStatus

    def __repr__(self):
        return create_repr(
            self, date=self.date, weekday=self.weekday, status=self.status
        )


@dataclass(frozen=True)
class PartialSubject:
    """
    Represents a partial subject.

    :param code: The subject code.
    :param description: The subject description.
    """

    code: str
    description: Optional[str]


@dataclass(frozen=True)
class Event:
    """
    Represents an agenda event.

    :param id: The ID of the event.
    :param type: The kind of event (annotation, homework, or classroom reservation).
    :param start: The start date and time.
    :param end: The end date and time.
    :param full_day: Whether the event is full-day or not.
    :param author: The author (a teacher in most cases) of the event.
    :param class_desc: The description of the classroom.
    :param notes: The actual text of the event.
    :param homework: The homework ID, if homework is provided.
    :param homework_item: The homework item, if homework is provided.
    :param subject: The subject of the event.
    """

    id: int
    type: EventCode
    start: datetime
    end: datetime
    full_day: bool
    author: str
    class_desc: str
    notes: Optional[str] = None
    homework: Optional[int] = None
    homework_item: Optional[List[str]] = None
    subject: Optional[PartialSubject] = None

    def __int__(self):
        return self.id

    def __str__(self):
        return self.class_desc

    def __repr__(self):
        return create_repr(
            self,
            id=self.id,
            type=self.type,
            start=self.start,
            end=self.end,
            author=self.author,
        )


@dataclass(frozen=True)
class AgendaDay:
    """
    Represents a day for the agenda.

    :param date: The date.
    :param events: The events of the day.
    """

    date: date
    events: List[Event]


@dataclass(frozen=True)
class Lesson:
    """
    Represents a lesson.

    :param id: The ID of the lesson.
    :param date: The date of the lesson.
    :param type: The type of lesson (normal or co-presence with and without support).
    :param position: The position of the lesson in that day.
    :param duration: The duration of the lesson.
    :param class_desc: The description of the classroom.
    :param subject: The subject of the lesson.
    :param status: The status of the lesson.
    """

    id: int
    date: date
    type: LessonEvent
    position: int
    duration: int
    class_desc: str
    subject: PartialSubject
    status: Optional[LessonStatus] = None

    def __int__(self):
        return self.id

    def __str__(self):
        return self.subject.description

    def __repr__(self):
        return create_repr(
            self,
            id=self.id,
            date=self.date,
            type=self.type,
            subject=self.subject,
            status=self.status,
        )


@dataclass(frozen=True)
class MIURData:
    """
    Represents data provided from the MIUR.

    :param code: The school code.
    :param division: The school division code.
    """

    code: str
    division: str


@dataclass(frozen=True)
class School:
    """
    Represents a school.

    :param code: The school code.
    :param name: The school name.
    :param dedication: The dedication of the school.
    :param city: The city of the school.
    :param province: The province of the school.
    :param miur: The school's data from MIUR.
    """

    code: str
    name: str
    dedication: str
    city: str
    province: str
    miur_data: MIURData

    def __str__(self):
        return self.name

    def __repr__(self):
        return create_repr(
            self, code=self.code, name=self.name, city=self.city, province=self.province
        )


@dataclass(frozen=True)
class Note:
    """
    Represents a school note.

    :param id: The ID of the note.
    :param type: The type of the note (annotation, disciplinary, warning and sanction).
    :param date: The day the note has been created.
    :param text: The text of the note.
    :param read: Whether the note has been read.
    :param author_name: The name of the author of the note.
    :param end: Optional. The day the disciplinary sanction ends. Only available when :attr:`type` is `sanction`.
    """

    id: int
    type: NoteType
    end: Optional[date]
    date: date
    text: str
    read: bool
    author_name: str

    def __int__(self):
        return self.id

    def __str__(self):
        return self.text

    def __repr__(self):
        return create_repr(
            self, id=self.id, date=self.date, type=self.type, read=self.read
        )


@dataclass(frozen=True)
class Teacher:
    """
    Represents a teacher.

    :param id: The ID of the teacher.
    :param name: The name of the teacher.
    """

    id: str
    name: str

    def __int__(self):
        return self.id

    def __str__(self):
        return self.name

    def __repr__(self):
        return create_repr(self, id=self.id, name=self.name)


@dataclass(frozen=True)
class Subject:
    """
    Represents a subject.

    :param id: The ID of the subject.
    :param description: The description of the subject.
    :param order: The position of the subject for a UI.
    :param teachers: The teachers that teach the subject.
    """

    id: int
    description: str
    order: int
    teachers: List[Teacher]
    grades: Optional[List["Grade"]]

    def __int__(self):
        return self.id

    def __str__(self):
        return self.description

    def __repr__(self):
        return create_repr(
            self,
            id=self.id,
            description=self.description,
            teachers=self.teachers if len(self.teachers) > 1 else None,
        )


@dataclass(frozen=True)
class Grade:
    """
    Represents a subject grade.

    :param subject: The subject of the grade.
    :param subject_code: The subject's shorter name.
    :param id: The ID of the grade.
    :param code: The kind of grade.
    :param date: The day the grade has been given.
    :param value: The actual value of the grade. (e.g. 6.25)
    :param display_value: The display value of the grade. (e.g. 6.25 is displayed as 6+)
    :param position: The position of the grade.
    :param family_notes: The notes of the grade visible to family and students.
    :param color: The color of the grade.
    :param canceled: Whether the grade has been canceled.
    :param underlined: Whether the grade is underlined.
    :param period: The period of the grade.
    :param component_position: The position of the component.
    :param component_description: The description of the component.
    :param weight: The weight of the grade.
    :param grade_master_id: The ID of the grade master.
    :param skill_id: The ID of the skill.
    :param skill_description: The description of the skill.
    :param skill_code: The code of the skill.
    :param skill_master_id: The ID of the skill master.
    """

    subject: Subject
    subject_code: str
    id: int
    code: GradeCode
    date: date
    value: Optional[int]
    display_value: str
    position: int
    family_notes: str
    color: str
    canceled: bool
    underlined: bool
    period: "Period"
    component_position: int
    component_description: str
    weight: int
    grade_master_id: int
    skill_id: int
    skill_description: str
    skill_code: str
    skill_master_id: int

    def __str__(self):
        return self.display_value

    def __int__(self):
        return self.value

    def __repr__(self):
        return create_repr(
            self,
            id=self.id,
            subject=self.subject.description,
            value=self.value or self.display_value,
            date=self.date,
        )


@dataclass(frozen=True)
class Day(SchoolDay):
    """
    Represents a full overview of a day.

    :param date: The date.
    :param weekday: The day weekday.
    :param status: The kind of day (normal, non-working, etc.).
    :param lessons: The lessons of that day.
    :param agenda: The events of that day.
    :param events: The absences of that day.
    :param grades: The grades of that day.
    :param notes: The notes of that day.
    """

    date: date
    lessons: List[Lesson]
    agenda: List[Event]
    events: List[AbsenceDay]
    grades: List[Grade]
    notes: List[Note]

    def __repr__(self) -> str:
        return create_repr(
            self,
            date=self.date,
            weekday=self.weekday,
            lessons=self.lessons,
            agenda=self.agenda,
            events=self.events,
            grades=self.grades,
            notes=self.notes,
        )


# This has been imported here to avoid circular imports
from .helpers.calendar.period import Period



"""
The modules are used to make manual HTTP requests to the Classeviva APIs. These come from the official API documentation and are implemented in the wrapper exactly as they were written on there.

Modules are useful when:

* You use the :attr:`~aiocvv.client.ClassevivaClient.me` attribute in :class:`~aiocvv.client.ClassevivaClient` to get the information you need more easily, which automatically uses the right module;
* You want to make a request to an endpoint that hasn't been/is partially implemented in this wrapper;
* You want to have more control over the request or response;
* You don't want to manually write every URL, like you would do with the :meth:`~aiocvv.client.ClassevivaClient.request` method.

.. note::
    Of course, if you decide to use modules,
    you will have to parse the response yourself,
    as they only return the raw response.

.. warning::
    The modules are not made to be manually constructed, but
    to be used through the :class:`~aiocvv.client.ClassevivaClient` instance.
"""

from .teachers import TeachersModule
from .students import StudentsModule
from .parents import ParentsModule
